/*
  Copyright (c) 2003-2004 The Regents of the University of California.
  All rights reserved.

  Permission is hereby granted, without written agreement and without
  license or royalty fees, to use, copy, modify, and distribute this
  software and its documentation for any purpose, provided that the
  above copyright notice and the following two paragraphs appear in all
  copies of this software and that appropriate acknowledgments are made
  to the research of the Metropolis group.

  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
  FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
  ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
  THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.

  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
  PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
  CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
  ENHANCEMENTS, OR MODIFICATIONS.


  METROPOLIS_COPYRIGHT_VERSION_1
  COPYRIGHTENDKEY

*/
#include "programcounter.h"
#include "event.h"
#include "scoreboard.h"
#include "GlobalTime.h"
#include "assert.h"
#include <algorithm>

String *ProgramCounter::NO_EVENT_TAG = new String("");
String *ProgramCounter::BEG_TAG = new String("begin");
String *ProgramCounter::END_TAG = new String("end");

ProgramCounter::ProgramCounter(process *pp): p(pp) {
    obj.object = (sc_object *) pp;
    obj.intfcName = new String("");
    _pScoreboard = NULL;
    stLabel = new String("");
    SchedState = SchedStateVal::RUN;
    cond = NULL;
    selected = -1;
    funcType = 0;
    awaitTestDepth = 0;
    mode = RunType::EXEC;
    _inSynch = false;
    _potentialUsedMedia = new MediaSet_t;
    currentLineNumber = (String*)0;
    eventTag = ProgramCounter::NO_EVENT_TAG;
    nextLineContextStack = new _2DStrListStack_t();
}

void ProgramCounter::initAwait(int numCs, bool nondet) {
    _csOrder.clear();
    for (int j = 0; j < numCs; j++) _csOrder.push_back(j);



    if (nondet) {
        for (int j = 0; j < numCs; j++) {
            int pos = (int)(random() % numCs);
            int swap = _csOrder[pos];
            _csOrder[pos] = _csOrder[j];
            _csOrder[j] = swap;
        } // end for
    } // end if
} // end initAwait

void ProgramCounter::addPendingEvent(event *e, int csNumber) {
    if (!e) return;
    _pendingEvents.insert(eventMap_t::value_type(e, csNumber));
} // end addPendingEvent

void ProgramCounter::freeAllPendingEvents() {
    _pendingEvents.clear();
} // end addPendingEvent

void ProgramCounter::savePendingEvents() {
    _oldPendingEvents = _pendingEvents;
} // end savePendingEvents

void ProgramCounter::loadPendingEvents() {
    _pendingEvents = _oldPendingEvents;
} // end loadPendingEvents

bool ProgramCounter::containsPendingEvent(event *e) {
    eventMap_t::iterator it = _pendingEvents.find(e);
    return (it != _pendingEvents.end());
} // end containsPendingEvent

bool ProgramCounter::setMustDo(event *e) {
    eventMap_t::iterator it = _pendingEvents.find(e);
    int csNumber = 0;
    if (it == _pendingEvents.end()) return false;
    csNumber = it->second;
    freeAllPendingEvents();
    addPendingEvent(e, csNumber);
    return true;
} // end setMustDo

bool ProgramCounter::setMustNotDo(event *e) {
    eventMap_t::iterator it = _pendingEvents.find(e);
    if (it == _pendingEvents.end()) return false;

    if (SchedState == SchedStateVal::RUN && selected == it->second)
        if (_sb.testSchedFlag) SchedState = SchedStateVal::TESTSCHEDULING;
        else  SchedState = SchedStateVal::UNKNOW;
    _pendingEvents.erase(it);
    selected = -1;
    return true;
} // end setMustNotDo

void ProgramCounter::setState(sc_object * pp, String *intfc, String *ll, int ss) {
    obj.object = pp;
    if (intfc!=Keep_Unchanged_String) obj.intfcName = (*intfc);
    if (ll!=Keep_Unchanged_String) stLabel = (*ll);
    SchedState = ss;
}

void ProgramCounter::saveState() {
    labelList.insert(labelList.begin(), new String(stLabel));
    nodeList.insert(nodeList.begin(), new node(obj));
    SchedStateList.insert(SchedStateList.begin(), new int(SchedState));
    funcTypeList.insert(funcTypeList.begin(), new int(funcType));

    _pendingEventsStack.insert(_pendingEventsStack.begin(), _pendingEvents);
    _pendingEvents.clear();

    _potentialUsedMediaList.push_back(_potentialUsedMedia);
    _potentialUsedMedia = new MediaSet_t;

    pushNewNextLineContext();
} // end saveState

void ProgramCounter::restoreState() {
    node * n = nodeList.front();
    String * l = labelList.front();
    int * s = SchedStateList.front();
    int * f = funcTypeList.front();

    popNextLineContext();

    nodeList.erase(nodeList.begin());
    labelList.erase(labelList.begin());
    SchedStateList.erase(SchedStateList.begin());
    funcTypeList.erase(funcTypeList.begin());
    obj = n;
    stLabel = l->c_str();
    SchedState = *s;
    funcType = *f;
    delete n;
    delete l;
    delete s;
    delete f;

    if (!_pendingEventsStack.empty()) {
        _pendingEvents = _pendingEventsStack.front();
        _pendingEventsStack.erase(_pendingEventsStack.begin());
    } // end if

    delete _potentialUsedMedia;
    _potentialUsedMedia = _potentialUsedMediaList.back();
    _potentialUsedMediaList.pop_back();
} // end restoreState

void ProgramCounter::setPotentialUsedMedia() {
    for (MediaSet_t::iterator it=_potentialUsedMedia->begin(); it!=_potentialUsedMedia->end(); it++) {
        medium * m = (medium*) (*it);
        m->_potentialUserPCs.insert(this);
    }
} // end setPotentialUsedMedia

void ProgramCounter::resetPotentialUsedMedia() {
    for (MediaSet_t::iterator it=_potentialUsedMedia->begin(); it!=_potentialUsedMedia->end(); it++) {
        medium * m = (medium*) (*it);
        m->_potentialUserPCs.erase(this);
    }
} // end resetPotentialUsedMedia

/**
 * Methods used for metamodel debugging:
 */
void ProgramCounter::enterAwaitTest() {
    awaitTestDepth++;
}

void ProgramCounter::exitAwaitTest() {
    awaitTestDepth--;
}

bool ProgramCounter::inAwaitTest() {
    return (awaitTestDepth > 0);
}

/**
 * Push the given single next-line string onto this PC's nextLineStack,
 * as a strList_t containing one element.  An empty string has special
 * meaning; if the string is empty, push an empty list.
 */
void ProgramCounter::pushNextLine(const char* nextLine) {
    strList_t* list = new strList_t();
    String* s = new String(nextLine);
    if (strcmp(nextLine, "") != 0) {
        list->push_back(s);
    }
    pushNextLineList(list);
}

/**
 * Add the given next-line string to the list on top of this PC's nextLineStack.
 */
void ProgramCounter::addNextLine(const char* nextLine) {
    strList_t* list = getNextLines();
    list->push_back(new String(nextLine));
}

/**
 * Push the given list of next-line Strings onto this PC's nextLineStack.
 */
void ProgramCounter::pushNextLineList(strList_t* list) {
    _2DStrList_t *nextLineStack = getCurrentNextLineContext();
    nextLineStack->push_back(list);
}

/**
 * Pop the top list of next-line Strings off this PC's nextLineStack,
 * and destroy it.
 */
void ProgramCounter::popNextLines() {
    void destroyListContents(strList_t*);
    _2DStrList_t *nextLineStack = getCurrentNextLineContext();
    assert(nextLineStack->size() > 0);
    strList_t* list;
    list = getNextLines();
    destroyListContents(list);
    nextLineStack->pop_back();
}

// /**
//  * Return the top list of next-line Strings currently on this PC's
//  * nextLineStack.
//  */
// strList_t* ProgramCounter::getNextLines() {
//     _2DStrList_t *nextLineStack = getCurrentNextLineContext();
//     int i = nextLineStack->size() - 1;
//     assert(i>=0);
//     return (nextLineStack->at(i));
// }

/**
 * Return the top non-empty list of next-line Strings currently on this PC's
 * nextLineStack, or null if there is no non-empty list.
 */
strList_t* ProgramCounter::getNextLines() {
    _2DStrList_t *nextLineStack;
    strList_t *list;
    for (int i = nextLineContextStack->size() - 1; i >= 0; i--) {
        nextLineStack = nextLineContextStack->at(i);
        for (int j = nextLineStack->size() - 1; j >= 0; j--) {
            list = nextLineStack->at(j);
            if (list->size() > 0) {
                return (list);
            }
        }
    }
    return ((strList_t *)0);
}

/**
 *
 */
void ProgramCounter::dumpNextLines() {
    _2DStrList_t *nextLineStack;
    strList_t *list;
    cout << "< ";
    for (int i = nextLineContextStack->size() - 1; i >= 0; i--) {
        nextLineStack = nextLineContextStack->at(i);
        cout << "{ ";
        for (int j = nextLineStack->size() - 1; j >= 0; j--) {
            list = nextLineStack->at(j);
            cout << "[ ";
            for (int k = 0; k < list->size(); k++) {
                String *s = list->at(k);
                if (k > 0) {
                    cout << ", ";
                }
                cout << s->c_str();
            }
            cout << " ]";
        }
        cout << " }";
    }
    cout << " >" << endl;
}

/**
 * Return the top next-line context on the next-line context stack.
 */
_2DStrList_t* ProgramCounter::getCurrentNextLineContext() {
    int i = nextLineContextStack->size() - 1;
    assert(i>=0);
    return (nextLineContextStack->at(i));
}

/**
 * Debugging version of getCurrentnextLineContext().
 */
void ProgramCounter::pushNewNextLineContext(const char* msg1, const char* msg2) {
    cout << msg1 << " -- " << msg2 << ": ";
    pushNewNextLineContext();
}

/**
 * Add a new next-line context stack to nextLineContextStack.
 */
void ProgramCounter::pushNewNextLineContext() {
    nextLineContextStack->push_back(new _2DStrList_t());
}

/**
 * Debugging version of popnextLineContext().
 */
void ProgramCounter::popNextLineContext(const char* msg1, const char* msg2) {
    cout << msg1 << " -- " << msg2 << ": ";
    popNextLineContext();
}

/**
 * Remove and destroy the top-most next-line context stack from
 * nextLineContextStack.
 */
void ProgramCounter::popNextLineContext() {
    void destroyListContents(_2DStrList_t*);
    _2DStrList_t* topStack = getCurrentNextLineContext();
    destroyListContents(topStack);
    nextLineContextStack->pop_back();
}

/**
 * Return the current currentLineNumber value.
 */
String* ProgramCounter::getCurrentLineNumber() {
    return currentLineNumber;
}

/**
 * Set this PC's currentLineNumber String to the one given.
 * As a side-effect, have the scoreboard update its list of all
 * processes' current-line strings, for access from the debugger.
 */
void ProgramCounter::setCurrentLineNumber(const char* lineNumber) {
    delete currentLineNumber;
    currentLineNumber = new String(lineNumber);
    getScoreboard()->updateCurrentLines();
}

/*
 * Memory management for String vectors, and String vector vectors,
 * and String vector vector vectors :-).
 */
void destroyListContents(strList_t *list) {
    if (list == (strList_t *)0) {
        return;
    }
    for (int i = 0; i < (int)(list->size()) - 1; i++) {
        String* element = list->at(i);
        delete element;
    }
    list->clear();
}

void destroyListContents(_2DStrList_t *list) {
    int i;
    for (i = 0; i < (int)(list->size()) - 1; i++) {
        strList_t* element = list->at(i);
        destroyListContents(element);
    }
    list->clear();
}

void destroyListContents(_2DStrListStack_t *list) {
    int i;
    for (i = 0; i < (int)(list->size()) - 1; i++) {
        _2DStrList_t* element = list->at(i);
        destroyListContents(element);
    }
    list->clear();
}
